import 'dart:async';

import 'package:flutter/foundation.dart';

/// Mutual exclusion.
/// Acquire and release lock
class Mutex {
  Completer<void>? _completer;

  /// Check if mutex is currently acquired
  bool get acquired => _completer != null;

  /// acquire while function is executing then release
  Future<T> lock<T>(Future<T> Function() function) => acquire().then((_) async {
        try {
          return await function();
        } finally {
          release();
        }
      });

  Future<void> acquire() {
    final c = _completer;
    if (c == null) {
      _completer = Completer();
      return SynchronousFuture(null);
    } else {
      return c.future.then((_) => acquire());
    }
  }

  void release() {
    final c = _completer;
    if (c == null) throw StateError('cannot release, nothing acquired');
    c.complete();
    _completer = null;
  }
}

/// Helper for running tasks with mutex
abstract class MutexRun {
  final Mutex mutex = Mutex();

  MutexRun();

  /// Run task exactly once, after first execution every other call will be ignored
  factory MutexRun.once() => _RunOnce();

  /// Run task, every call made while another call is running will be ignored until the first one completes
  factory MutexRun.first() => _RunFirst();

  /// Run task once every [delay], after the first call other call will be ignored until delay passed.
  factory MutexRun.every(Duration delay) => _RunEvery(delay);

  /// Run task after [delay], every call will invalidate the previous one and only the last call within the delay
  /// will be performed
  factory MutexRun.last(Duration delay) => _RunLast(delay);

  /// Execute async task (depending on rules) and return true if [function] has been executed
  Future<bool> call(Future Function() function);
}

class _RunOnce extends MutexRun {
  bool _done = false;

  @override
  Future<bool> call(Future Function() function) => _done || mutex.acquired
      ? SynchronousFuture(false)
      : mutex.lock<bool>(() async {
          if (_done) return false;
          await function();
          return _done = true;
        });
}

class _RunFirst extends MutexRun {
  @override
  Future<bool> call(Future Function() function) =>
      mutex.acquired ? SynchronousFuture(false) : mutex.lock(function).then((_) => true);
}

class _RunEvery extends MutexRun {
  final Duration delay;
  DateTime? _lastExecution;

  _RunEvery(this.delay) : assert(delay > Duration.zero, 'Use MutexRun.first instead');

  @override
  Future<bool> call(Future Function() function) => mutex.acquired
      ? SynchronousFuture(false)
      : mutex.lock<bool>(() async {
          if (_lastExecution?.isBefore(DateTime.now().subtract(delay)) ?? false) return false;
          await function();
          _lastExecution = DateTime.now();
          return true;
        });
}

class _RunLast extends MutexRun {
  final Duration delay;
  late DateTime _lastCall;

  _RunLast(this.delay) : assert(delay > Duration.zero, 'Use MutexRun.first instead');

  @override
  Future<bool> call(Future Function() function) {
    final now = _lastCall = DateTime.now();
    return Future.delayed(
        delay,
        () => mutex.acquired
            ? SynchronousFuture(false)
            : mutex.lock<bool>(() async {
                if (now != _lastCall) return false;
                await function();
                return true;
              }));
  }
}
